from llm_party.model.session_models import LLMAgent
from llm_party.attendee.base.base_trait import AttendeeTraitBase



class RawLLMAgent(LLMAgent, AttendeeTraitBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)  # Call the __init__ method of LLMAgent
        AttendeeTraitBase.__init__(self)  # Call the __init__ method of BaseAttendeeTrait



import yaml
from abc import ABC, abstractmethod, ABCMeta
from typing import Optional, List, Literal
from itertools import cycle
from llm_wrap.llm_response import get_response as get_ai_response
from llm_party.registry.attendee_registry import attendee_registry
from marshmallow import (
    Schema,
    fields,
    post_load,
    pre_load,
    ValidationError,
    validates_schema,
    EXCLUDE,
)

from marshmallow import (
    Schema,
    fields,
    post_load,
    ValidationError,
    validates_schema,
    EXCLUDE,
)
from datetime import datetime
from ..logger import get_logger
logging = get_logger()


class InstructionSchema(Schema):
    text = fields.Str(required=True)
    target = fields.Str(allow_none=True)
    version = fields.Str(allow_none=True)
    hash = fields.Str(allow_none=True)

    @post_load
    def make_instruction(self, data, **kwargs):
        return Instruction(**data)


class AttendeeSchema(Schema):
    type = fields.Str(required=True)
    name = fields.Str(required=True)
    role = fields.Str(required=True)
    instruction = fields.Nested(InstructionSchema, required=True)
    attendee_params = fields.Dict(required=True)  # allow_none=True)
    # Define the fields but exclude them by default
    llm_api = fields.Str(allow_none=True)
    llm_api_params = fields.Dict(allow_none=True)

    @pre_load
    def process_input(self, data, **kwargs):
        if attendee_registry.get(data["type"]):
            # No need to check for presence here, as they are optional and validated later
            pass
        else:
            # Remove LLMAgent-specific fields if present
            data.pop("llm_api", None)
            data.pop("llm_api_params", None)
        return data

    @post_load
    def make_attendee(self, data, **kwargs):
        attendee_type = data["type"]
        AttendeeClass = attendee_registry.get(attendee_type)
        if AttendeeClass is not None:
            return AttendeeClass(**data)
        else:
            raise ValidationError(f'The type must be a defined class and a subclass of Attendee. "{attendee_type}" does not meet these conditions.')


class MessageSchema(Schema):
    """
    Schema for serializing and deserializing Message objects.
    """

    sender = fields.Nested(AttendeeSchema, required=True)
    text = fields.Str(required=True)
    timestamp = fields.DateTime(format="%Y-%m-%dT%H:%M:%S", required=True)

    @post_load
    def make_message(self, data, **kwargs) -> "Message":
        """
        Create a Message object from deserialized data.

        Args:
            data (dict): Deserialized data.
            **kwargs: Additional keyword arguments.

        Returns:
            Message: A Message object.
        """
        return Message(**data)


class ChatSessionSchema(Schema):
    title = fields.Str(required=True)
    purpose = fields.Str(required=True)
    attendees = fields.List(fields.Nested(AttendeeSchema), required=True)
    settings = fields.Dict(keys=fields.Str(), values=fields.Raw(), required=True)
    start_at = fields.DateTime(format="%Y-%m-%dT%H:%M:%S", allow_none=True)
    updated_at = fields.DateTime(format="%Y-%m-%dT%H:%M:%S", allow_none=True)
    end_at = fields.DateTime(format="%Y-%m-%dT%H:%M:%S", allow_none=True)
    finish_reason = fields.Str(allow_none=True)
    status = fields.Str(required=True)
    chat_history = fields.List(fields.Nested(MessageSchema), allow_none=True)

    class Meta:
        unknown = EXCLUDE

    @post_load
    def make_chat_session(self, data, **kwargs):
        return ChatSession(**data)


class Instruction:
    def __init__(self, text, target=None, version=None, hash=None):
        self.update(text, target, version, hash)

    def update(self, text, target=None, version=None, hash=None):
        self.text = text
        self.target = target
        self.version = version
        self.hash = hash

class AttendeeMeta(ABCMeta):
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        if name != 'Attendee':  # Avoid registering the base Attendee class
            attendee_registry.register(cls)


class Attendee(ABC, metaclass=AttendeeMeta):
    def __init__(self, name, role, instruction, attendee_params=None):
        if not isinstance(instruction, Instruction):
            raise TypeError("instruction must be an instance of Instruction")
        self.name = name
        self.role = role
        self.instruction = instruction
        self.attendee_params = attendee_params if attendee_params is not None else {}
        self.type = self.__class__.__name__

    def send_new_message(self, session: "ChatSession") -> "Message":
        response = self.generate_response(session.chat_history)
        message = Message(self, response)
        session.chat_history.append(message)
        return message

    @abstractmethod
    def generate_response(self, chat_history: List["Message"]) -> str:
        pass


class Human(Attendee):
    def __init__(self, name, role, instruction, attendee_params=None, **kwargs):
        super().__init__(name, role, instruction, attendee_params)
        # You can either store the extra kwargs or simply ignore them
        # For example, if you want to store them:
        self.extra_data = kwargs

    def generate_response(self, chat_history):
        raise NotImplementedError(
            "The generate_response method is not implemented for Human attendees."
        )


# TODO Make this an abstract class
class LLMAgent(Attendee):
    def __init__(
        self,
        name,
        role,
        instruction,
        attendee_params=None,
        llm_api=None,
        llm_api_params=None,
        **kwargs,
    ):
        super().__init__(name, role, instruction, attendee_params)
        self.llm_api = llm_api
        self.llm_api_params = llm_api_params if llm_api_params is not None else {}
        # Again, store or ignore the extra kwargs
        self.extra_data = kwargs


    # # TODO: @abstractmethod
    # def trait_instruction(self) -> str:
    #     # TODO: replace it with NotImplementedError
    #     return ""

    def generate_response(self, chat_history: List["Message"]) -> str:
        if self.llm_api is None:
            raise ValueError(
                "LLMAgent 'llm_api' attribute is not set to a supported API identifier. Please see the documentation for supported APIs in `llm_wrap` library."
            )

        messages = self.convert_chat_history_to_openai_messages(chat_history)

        logging.debug(f"messages (no system): {messages[1:]}")
        try:
            response = get_ai_response(
                llm_api=self.llm_api, messages=messages, **self.llm_api_params
            )
        except ValueError as e:
            raise ValueError(f"Error generating response: {str(e)}")

        return response

    @property
    def initial_instruction(self):
        if hasattr(self, "trait_instruction"):
            return self.instruction.text + "\n" + self.trait_instruction()
        else:
            return self.instruction.text

    def convert_chat_history_to_openai_messages(self, chat_history: List["Message"]) -> List[dict]:
        # Create the initial system message
        messages = [{"role": "system", "content": self.initial_instruction}]

        # Define the roles for the chat history messages
        roles = cycle(["user", "assistant"])

        # Convert each message in the chat history to the required format
        for message in chat_history:
            formatted_message = {"role": next(roles), "content": message.text}
            messages.append(formatted_message)

        return messages


class Message:
    def __init__(self, sender: Attendee, text: str, timestamp: Optional[datetime]=None):
        self.sender = sender
        self.text = text
        self.timestamp = timestamp if timestamp else datetime.now()

    @staticmethod
    def create(
        sender_name: str,
        sender_role: str,
        sender_type: str,
        sender_attendee_params: dict,
        text: str,
        instruction_text: str = "",
        instruction_target: str = None,
        instruction_version: str = None,
        instruction_hash: str = None,
    ) -> "Message":
        """
        Create a Message object.

        Args:
            sender_name (str): The name of the sender.
            sender_role (str): The role of the sender.
            sender_type (str): The type of the sender.
            sender_attendee_params (dict): The attendee parameters of the sender.
            text (str): The text of the message.
            instruction_text (str): The text of the instruction.
            instruction_target (str): The target of the instruction.
            instruction_version (str): The version of the instruction.
            instruction_hash (str): The hash of the instruction.

        Returns:
            Message: A Message object.
        """
        if instruction_target is None:
            instruction_target = ""
        if instruction_version is None:
            instruction_version = ""
        if instruction_hash is None:
            instruction_hash = ""
        instruction = Instruction(
            text=instruction_text,
            target=instruction_target,
            version=instruction_version,
            hash=instruction_hash,
        )

        if sender_type == "Human":
            sender = Human(
                name=sender_name,
                role=sender_role,
                type=sender_type,
                instruction=instruction,
                attendee_params=sender_attendee_params,
            )
        elif sender_type == "LLMAgent":
            sender = LLMAgent(
                name=sender_name,
                role=sender_role,
                type=sender_type,
                instruction=instruction,
                attendee_params=sender_attendee_params,
            )
        else:
            raise ValueError(
                f"Invalid sender_type: {sender_type}. Expected 'Human' or 'LLMAgent'."
            )

        return Message(sender=sender, text=text, timestamp=datetime.now())

    def to_openai_message(self):
        return {"role": self.sender.role, "content": self.text}

    def to_dict(self):
        schema = MessageSchema()
        return schema.dump(self)

class ChatSession:
    def __init__(
        self,
        title: str=None,
        purpose: str=None,
        attendees: List[Attendee]=None,
        settings: dict=None,
        start_at: datetime=None,
        updated_at: datetime=None,
        end_at: datetime=None,
        finish_reason: str=None,
        chat_history: List[Message]=None,
        status: Literal["initialized", "started", "paused", "resumed", "completed_successfully", "completed_with_failure", "cancelled"]="initialized",
    ):
        self.title = title
        self.purpose = purpose
        self.attendees = attendees if attendees is not None else []
        self.settings = settings if settings is not None else {}
        self.chat_history = chat_history if chat_history is not None else []
        self.finish_reason = finish_reason
        self.start_at = start_at
        self.updated_at = updated_at
        self.end_at = end_at
        self.status = status

    def add_attendee(self, attendee):
        if attendee not in self.attendees:
            self.attendees.append(attendee)
            self.updated_at = datetime.now()
        else:
            raise ValueError(f"Attendee {attendee.name} is already part of the session")

    def remove_attendee(self, attendee):
        try:
            self.attendees.remove(attendee)
            self.updated_at = datetime.now()
        except ValueError:
            raise ValueError(f"Attendee {attendee.name} is not part of the session")

    def configure(self, settings):
        self.settings = settings

    def start(self):
        self.start_at = datetime.now()
        self.updated_at = self.start_at
        self.status = "started"

    def end(self, status, finish_reason):
        self.status = status
        self.finish_reason = finish_reason
        self.end_at = datetime.now()

    def set_purpose(self, purpose):
        self.purpose = purpose

    def get_purpose(self):
        return self.purpose

    @staticmethod
    def from_json(json_string):
        schema = ChatSessionSchema()
        try:
            # Deserialize the JSON string into a dictionary
            data = schema.loads(json_string)
            # Check if data is already a ChatSession instance
            if isinstance(data, ChatSession):
                return data
            # Otherwise, create a new ChatSession instance using the deserialized data
            return ChatSession(**data)
        except ValidationError as err:
            raise ValueError(f"Invalid JSON data: {err.messages}")

    def to_json(self, pretty_print=False):
        """
        Serializes the ChatSession object to JSON.

        Returns:
            str: The serialized JSON representation of the ChatSession object.

        Raises:
            ValueError: If there is an error serializing the ChatSession object.
        """
        schema = ChatSessionSchema()
        try:
            if pretty_print:
                return schema.dumps(self, indent=4)
            else:
                return schema.dumps(self)
        except ValidationError as err:
            raise ValueError(f"Error serializing ChatSession: {err.messages}")

    def to_dict(self):
        """
        Converts the ChatSession object to a dictionary.

        Returns:
            dict: The dictionary representation of the ChatSession object.

        Raises:
            ValueError: If there is an error converting the ChatSession object.
        """
        schema = ChatSessionSchema()
        try:
            json_data = schema.dumps(self)
            return yaml.safe_load(json_data)
        except ValidationError as err:
            raise ValueError(f"Error converting ChatSession to dict: {err.messages}")

    def to_yaml(self):
        """
        Serializes the ChatSession object to YAML.

        Returns:
            str: The serialized YAML representation of the ChatSession object.

        Raises:
            ValueError: If there is an error serializing the ChatSession object.
        """
        try:
            dict_data = self.to_dict()
            return yaml.dump(dict_data, default_flow_style=False)
        except ValidationError as err:
            raise ValueError(f"Error serializing ChatSession to YAML: {err.messages}")

    def get_message_from_history(self, n: int) -> dict:
        """
        Returns the dictionary representation of the n-th message in the chat history.

        Args:
            n (int): The index of the message in the chat history. You can specify a negative index to get the message from the end of the chat history.

        Returns:
            dict: The dictionary representation of the n-th message in the chat history. The keys of the dictionary corresponds to the fields defined in MessageSchema.
        """
        if n < len(self.chat_history):
            return self.chat_history[n].to_dict()
        else:
            raise IndexError("Message index out of range.")

    def set_max_messages(self, value: int) -> None:
        """
        Set the max number of messages in the session settings.

        Args:
            value (int): The value to set for the max number of messages.
        """
        self.settings["max_sent_message"] = value

    def increase_max_messages(self, increment: int) -> None:
        """
        Increase the max number of messages by a specified increment. This property is used to control the number of turns in the session between agents.

        Args:
            increment (int): The amount to increase the max number of messages by.
        """
        if "max_sent_message" in self.settings:
            self.settings["max_sent_message"] += increment
        else:
            raise KeyError("The 'max_sent_message' setting does not exist.")

    @validates_schema
    def validate_chat_session(self, data, **kwargs):
        # Custom validation logic here
        pass



import os
import yaml
from datetime import datetime
import json
import unittest
from llm_party.service import start_session
from llm_party.service.session_service import validate_config_file, initialize_chat_session, progress_chat_session, create_first_message, start_session_with_user_first_message
from llm_party.model.session_models import ChatSession, MessageSchema, LLMAgent, Human, Message
from unittest.mock import patch, mock_open
from unittest import skipIf
original_open = open


USE_ACTUAL_LLM_API = False

def skip_if_no_api():
    return skipIf(not USE_ACTUAL_LLM_API, "Skipping this test because USE_ACTUAL_LLM_API is False")

def saved_dir():
    return './tests/service/data/saved/'


def _get_valid_config_file_path():
    return f"{service_data_path()}../../data/valid_config.json"

def _get_invalid_config_file_path():
    return f"{service_data_path()}invalid_config.json"

def service_data_path():
    return './tests/service/data/'

class TestConfigFileValidation(unittest.TestCase):
    def setUp(self):
        self.valid_config_file = _get_valid_config_file_path()
        self.invalid_config_file = _get_invalid_config_file_path()
        self.missing_fields_file1 = f"{service_data_path()}missing_fields1.json"
        self.missing_fields_file2 = f"{service_data_path()}missing_fields2.json"
        self.extra_fields_file = f"{service_data_path()}extra_fields.json"
        self.wrong_type_fields_file1 = f"{service_data_path()}wrong_type_fields1.json"
        self.wrong_type_fields_file2 = f"{service_data_path()}wrong_type_fields2.json"
        self.invalid_json_format_file = f"{service_data_path()}invalid_json_format_file.json"

    def test_valid_config_file(self):
        try:
            validate_config_file(self.valid_config_file)
        except ValueError:
            self.fail("validate_config_file() raised ValueError unexpectedly!")

    def test_invalid_config_file(self):
        with self.assertRaises(ValueError) as context:
            validate_config_file(self.invalid_config_file)
        self.assertIn("Error reading the configuration file", str(context.exception))

    def test_missing_fields_file1(self):
        with self.assertRaises(ValueError) as context:
            validate_config_file(self.missing_fields_file1)
        self.assertIn("Each LLM agent must have 'llm_api_params' field", str(context.exception))

    def test_missing_fields_file2(self):
        with self.assertRaises(ValueError) as context:
            validate_config_file(self.missing_fields_file2)
        self.assertIn("The configuration file must contain more than two agents", str(context.exception))

    def test_wrong_type_fields_file1(self):
        with self.assertRaises(ValueError) as context:
            validate_config_file(self.wrong_type_fields_file1)
        self.assertIn("'max_sent_message' must be an integer", str(context.exception))

    def test_wrong_type_fields_file2(self):
        with self.assertRaises(ValueError) as context:
            validate_config_file(self.wrong_type_fields_file2)
        self.assertIn("'max_sent_message' must be greater than 0", str(context.exception))

    def test_invalid_json_file(self):
        with self.assertRaises(ValueError) as context:
            validate_config_file(self.invalid_json_format_file)
        self.assertIn("Error reading the configuration file", str(context.exception))

    def test_file_not_exists(self):
        with self.assertRaises(ValueError) as context:
            validate_config_file("not_exists.json")
        self.assertIn("The configuration file not_exists.json does not exist or is not accessible", str(context.exception))

class TestInitializeChatSession(unittest.TestCase):
    def setUp(self):
        self.valid_config_file = _get_valid_config_file_path()
        self.invalid_config_file = _get_invalid_config_file_path()
        self.invalid_schema_config_file = 'invalid_schema_config.json'

    def test_valid_config_file(self):
        try:
            chat_session = initialize_chat_session(self.valid_config_file)
            self.assertIsInstance(chat_session, ChatSession)
        except ValueError:
            self.fail("initialize_chat_session() raised ValueError unexpectedly!")

    def test_invalid_config_file(self):
        with self.assertRaises(ValueError):
            initialize_chat_session(self.invalid_config_file)

    def test_invalid_schema_config_file(self):
        with self.assertRaises(ValueError):
            initialize_chat_session(self.invalid_schema_config_file)


class TestProgressChatSession(unittest.TestCase):
    def setUp(self):
        self.valid_config_file = _get_valid_config_file_path()

    @skip_if_no_api()
    def test_progress_chat_session_success(self):
        # Initialize a chat session with a valid configuration
        chat_session = initialize_chat_session(self.valid_config_file)
        # Progress the chat session
        chat_session, _ = progress_chat_session(chat_session, save_dir=saved_dir())
        # Check that the chat session status and finish reason are set correctly
        self.assertEqual(chat_session.status, 'completed_successfully')
        self.assertEqual(chat_session.finish_reason, 'MaxSentMessage')
        # Check that the chat history is updated correctly
        self.assertEqual(len(chat_session.chat_history), chat_session.settings['max_sent_message'])
        # Check that the end_at field is set
        self.assertIsNotNone(chat_session.end_at)

    @patch('llm_party.model.session_models.get_ai_response', return_value='Mocked response')
    def test_progress_chat_session_with_mocked_api(self, mock_get_ai_response):
        # Initialize a chat session with a valid configuration
        chat_session = initialize_chat_session(self.valid_config_file)
        # Progress the chat session
        chat_session, _ = progress_chat_session(chat_session, save_dir=saved_dir())
        # Check that the chat session status and finish reason are set correctly
        self.assertEqual(chat_session.status, 'completed_successfully')
        self.assertEqual(chat_session.finish_reason, 'MaxSentMessage')
        # Check that the chat history is updated correctly
        self.assertEqual(len(chat_session.chat_history), chat_session.settings['max_sent_message'])
        # Check that the end_at field is set
        self.assertIsNotNone(chat_session.end_at)
        # Check that the get_ai_response function was called the expected number of times
        self.assertEqual(mock_get_ai_response.call_count, chat_session.settings['max_sent_message'])

    @patch('llm_party.model.session_models.Attendee.send_new_message', side_effect=Exception('Test exception'))
    def test_progress_chat_session_failure(self, mock_send_new_message):
        # Initialize a chat session with a valid configuration
        chat_session = initialize_chat_session(self.valid_config_file)
        # Progress the chat session
        try:
            chat_session, _ = progress_chat_session(chat_session, save_dir=saved_dir())
        except Exception:
            pass
        # Check that the chat session status and finish reason are set correctly
        self.assertEqual(chat_session.status, 'completed_with_failure')
        self.assertTrue(chat_session.finish_reason.startswith('LLMAPIFailure'))

    @skip_if_no_api()
    def test_chat_session_completion_success(self):
        # Initialize a chat session with a valid configuration
        chat_session = initialize_chat_session(self.valid_config_file)
        # Progress the chat session
        chat_session, save_file_path = progress_chat_session(chat_session, save_dir=saved_dir())
        # Check that the chat session status and finish reason are set correctly
        self.assertEqual(chat_session.status, 'completed_successfully')
        self.assertEqual(chat_session.finish_reason, 'MaxSentMessage')
        # Check that the ChatSession object is correctly serialized to a JSON file
        with open(save_file_path, 'r') as f:
            saved_chat_session = json.load(f)
        self.assertEqual(saved_chat_session, chat_session.to_json())

    @patch('llm_party.model.session_models.get_ai_response', return_value='Test response')
    def test_chat_session_completion_success_with_mock(self, mock_get_ai_response):
        # Initialize a chat session with a valid configuration
        chat_session = initialize_chat_session(self.valid_config_file)
        # Progress the chat session
        chat_session, save_file_path = progress_chat_session(chat_session, save_dir=saved_dir())
        # Check that the chat session status and finish reason are set correctly
        self.assertEqual(chat_session.status, 'completed_successfully')
        self.assertEqual(chat_session.finish_reason, 'MaxSentMessage')
        # Check that the ChatSession object is correctly serialized to a JSON file
        with open(save_file_path, 'r') as f:
            saved_chat_session = json.load(f)
        self.assertEqual(saved_chat_session, json.loads(chat_session.to_json()))
        # Check that the get_ai_response function was called
        mock_get_ai_response.assert_called()

    @patch('llm_party.model.session_models.Attendee.send_new_message', side_effect=Exception('Test exception'))
    def test_chat_session_completion_failure(self, mock_send_new_message):
        # Initialize a chat session with a valid configuration
        chat_session = initialize_chat_session(self.valid_config_file)
        # Progress the chat session
        try:
            chat_session, _ = progress_chat_session(chat_session, save_dir=saved_dir())
        except Exception:
            pass
        # Check that the chat session status and finish reason are set correctly
        self.assertEqual(chat_session.status, 'completed_with_failure')
        self.assertTrue(chat_session.finish_reason.startswith('LLMAPIFailure'))

    @skip_if_no_api()
    def test_chat_session_completion_save_file_path_not_writable(self):
        def side_effect(*args, **kwargs):
            if args[0] == self.valid_config_file:
                return original_open(*args, **kwargs)
            else:
                raise PermissionError

        with patch('builtins.open', side_effect=side_effect) as mock_open:
            # Initialize a chat session with a valid configuration
            chat_session = initialize_chat_session(self.valid_config_file)
            # Progress the chat session
            with self.assertRaises(PermissionError):
                chat_session, _ = progress_chat_session(chat_session, save_dir=saved_dir())

    def side_effect_os_error(*args, **kwargs):
        if 'w' in args[1]:
            raise OSError
        return original_open(*args, **kwargs)

    @skip_if_no_api()
    @patch('builtins.open', side_effect=side_effect_os_error)
    def test_chat_session_completion_not_enough_disk_space(self, mock_open):
        # Initialize a chat session with a valid configuration
        chat_session = initialize_chat_session(self.valid_config_file)
        # Progress the chat session
        with self.assertRaises(OSError):
            chat_session, _ = progress_chat_session(chat_session, save_dir=saved_dir())

    @patch('llm_party.model.session_models.get_ai_response', return_value='Test response')
    def test_chat_session_completion_not_enough_disk_space_with_mock(self, mock_get_ai_response):
        def side_effect_os_error(*args, **kwargs):
            if 'w' in args[1]:
                raise OSError
            return original_open(*args, **kwargs)

        with patch('builtins.open', side_effect=side_effect_os_error):
            # Initialize a chat session with a valid configuration
            chat_session = initialize_chat_session(self.valid_config_file)
            # Progress the chat session
            with self.assertRaises(OSError):
                chat_session, _ = progress_chat_session(chat_session, save_dir=saved_dir())

    @patch('llm_party.model.session_models.get_ai_response', return_value='Test response')
    def test_chat_session_completion_save_file_path_not_writable_with_mock(self, mock_get_ai_response):
        def side_effect(*args, **kwargs):
            if args[0] == self.valid_config_file:
                return original_open(*args, **kwargs)
            else:
                raise PermissionError

        with patch('builtins.open', side_effect=side_effect):
            # Initialize a chat session with a valid configuration
            chat_session = initialize_chat_session(self.valid_config_file)
            # Progress the chat session
            with self.assertRaises(PermissionError):
                chat_session, _ = progress_chat_session(chat_session, save_dir=saved_dir())


class TestStartSession(unittest.TestCase):
    def setUp(self):
        self.valid_config_file_path = _get_valid_config_file_path()
        self.invalid_config_file_path = _get_invalid_config_file_path()
        self.save_dir = saved_dir()
        self.save_file_name = "test_chat_session.json"

    @skip_if_no_api()
    def test_start_session_valid_config(self):
        chat_session, saved_file_path = start_session(self.valid_config_file_path, save_dir=self.save_dir, save_file_name=self.save_file_name)
        self.assertIsInstance(chat_session, ChatSession)
        self.assertEqual(chat_session.status, 'completed_successfully')
        self.assertEqual(saved_file_path, os.path.join(self.save_dir, self.save_file_name))

    def test_start_session_invalid_config(self):
        # Test start_session with an invalid configuration file
        with self.assertRaises(ValueError):
            start_session(self.invalid_config_file_path, save_dir=self.save_dir, save_file_name=self.save_file_name)

    @patch('llm_party.service.session_service.progress_chat_session', side_effect=Exception('Test exception'))
    def test_start_session_failed_chat(self, mock_progress_chat_session):
        # Test start_session with a valid configuration file that causes the chat to fail
        with self.assertRaises(Exception) as context:
            start_session(self.valid_config_file_path, save_dir=self.save_dir, save_file_name=self.save_file_name)
        self.assertEqual(str(context.exception), 'Test exception')

    @patch('llm_party.model.session_models.get_ai_response', return_value='Mocked response')
    def test_start_session_valid_config_with_mock(self, mock_get_ai_response):
        # Test start_session with a valid configuration file
        chat_session, saved_file_path = start_session(self.valid_config_file_path, save_dir=self.save_dir, save_file_name=self.save_file_name)
        self.assertIsInstance(chat_session, ChatSession)
        self.assertEqual(chat_session.status, 'completed_successfully')
        self.assertEqual(saved_file_path, os.path.join(self.save_dir, self.save_file_name))

    @patch('llm_party.model.session_models.get_ai_response', return_value='Mocked response')
    def test_start_session_invalid_config_with_mock(self, mock_get_ai_response):
        # Test start_session with an invalid configuration file
        with self.assertRaises(ValueError):
            start_session(self.invalid_config_file_path, save_dir=self.save_dir, save_file_name=self.save_file_name)

    @patch('llm_party.service.session_service.progress_chat_session', side_effect=Exception('Test exception'))
    def test_start_session_failed_chat_with_mock(self, mock_progress_chat_session):
        # Test start_session with a valid configuration file that causes the chat to fail
        with self.assertRaises(Exception) as context:
            start_session(self.valid_config_file_path, save_dir=self.save_dir, save_file_name=self.save_file_name)
        self.assertEqual(str(context.exception), 'Test exception')

    @patch('llm_party.model.session_models.get_ai_response', return_value='Mocked response')
    def test_start_session_save_as_yaml(self, mock_get_ai_response):
        # Test start_session with save_yaml=True
        chat_session, saved_file_path = start_session(
            self.valid_config_file_path, 
            save_dir=self.save_dir, 
            save_file_name=self.save_file_name, 
            save_yaml=True
        )
        self.assertIsInstance(chat_session, ChatSession)
        self.assertEqual(chat_session.status, 'completed_successfully')
        self.assertEqual(saved_file_path, os.path.join(self.save_dir, self.save_file_name))
        # Check if the saved file is a valid YAML file
        with open(saved_file_path, 'r') as f:
            try:
                yaml.safe_load(f)
            except yaml.YAMLError:
                self.fail("The saved file is not a valid YAML file.")

    def tearDown(self):
        # Teardown code here
        pass


class TestSessionService(unittest.TestCase):
    def test_create_first_message(self):
        # Test with a Human sender
        sender_name = "Test Human"
        sender_role = "user"
        sender_type = "Human"
        sender_attendee_params = {}
        text = "Hello, world!"
        serialized_message = create_first_message(sender_name, sender_role, sender_type, sender_attendee_params, text)
        message_schema = MessageSchema()
        message = message_schema.load(serialized_message)
        self.assertIsInstance(message.sender, Human)
        self.assertEqual(message.text, text)

        # Test with an LLMAgent sender
        sender_name = "Test LLMAgent"
        sender_role = "assistant"
        sender_type = "LLMAgent"
        sender_attendee_params = {"llm_api": "openai"}
        text = "Hello, user!"
        serialized_message = create_first_message(sender_name, sender_role, sender_type, sender_attendee_params, text)
        message = message_schema.load(serialized_message)
        self.assertIsInstance(message.sender, LLMAgent)
        self.assertEqual(message.text, text)

        # Test with an invalid sender_type
        sender_type = "InvalidType"
        with self.assertRaises(ValueError):
            serialized_message = create_first_message(sender_name, sender_role, sender_type, sender_attendee_params, text)

    @patch('llm_party.model.session_models.get_ai_response', return_value='Mocked response')
    def test_start_session_with_user_first_message_with_mock(self, mock_get_ai_response):
        # Test with a Human sender
        config_file_path = _get_valid_config_file_path()
        first_message = "Hello, world!"
        sender_name = "Test Human"
        sender_role = "user"
        sender_type = "Human"
        sender_attendee_params = {}
        save_dir = None
        save_file_name = None

        chat_session, saved_file_path = start_session_with_user_first_message(
            config_file_path, first_message, sender_name, sender_role, sender_type, sender_attendee_params, save_dir, save_file_name
        )

        self.assertIsInstance(chat_session, ChatSession)
        self.assertEqual(len(chat_session.chat_history), 2)
        self.assertIsInstance(chat_session.chat_history[0], Message)
        self.assertEqual(chat_session.chat_history[0].text, first_message)

    @skip_if_no_api()
    def test_start_session_with_user_first_message(self, mock_get_ai_response):
        # Test with a Human sender
        config_file_path = _get_valid_config_file_path()
        first_message = "Hello, world!"
        sender_name = "Test Human"
        sender_role = "user"
        sender_type = "Human"
        sender_attendee_params = {}
        save_dir = None
        save_file_name = None

        chat_session, saved_file_path = start_session_with_user_first_message(
            config_file_path, first_message, sender_name, sender_role, sender_type, sender_attendee_params, save_dir, save_file_name
        )

        self.assertIsInstance(chat_session, ChatSession)
        self.assertEqual(len(chat_session.chat_history), 2)
        self.assertIsInstance(chat_session.chat_history[0], Message)
        self.assertEqual(chat_session.chat_history[0].text, first_message)


if __name__ == "__main__":
    unittest.main()


from llm_party.attendee.default.default_attendee_trait import FacilitatorTrait, CommunicatorTrait
from llm_party import RawLLMAgent


class FacilitatorAgent(RawLLMAgent):
    character_classes = [FacilitatorTrait, CommunicatorTrait]


class AttendeeRegistry:
    def __init__(self):
        self._registry = {}

    def register(self, attendee_class):
        self._registry[attendee_class.__name__] = attendee_class

    def get(self, attendee_type):
        return self._registry.get(attendee_type)

# Create a global instance of the registry
attendee_registry = AttendeeRegistry()


import yaml
from typing import Any

import yaml
from typing import Any, List

def _get_base_dir() -> str:
    import os
    return os.path.dirname(os.path.abspath(__file__)) + "/../.."

class InstructionRegistry:
    def __init__(self, yaml_file_paths: List[str]):
        self._registry = {}
        for yaml_file_path in yaml_file_paths:
            with open(f"{_get_base_dir()}/{yaml_file_path}", 'r') as file:
                self._registry.update(yaml.safe_load(file))

    def get(self, str_name: str) -> Any:
        return self._registry.get(str_name)

    def __getattr__(self, name: str) -> Any:
        if name in self._registry:
            return self._registry[name]
        raise AttributeError(f"No such attribute: {name}")

    def __iter__(self):
        return iter(self._registry)

# Create a global instance of the registry
instruction_registry = InstructionRegistry(['llm_party/attendee/default/default_instruction.yaml', 'llm_party/attendee/custom/custom_instruction.yaml'])




def create_session(llm_api='openai', **kwargs):
    """Creates a new session.

    Args:
        llm_api (str, optional): _description_. Defaults to 'openai'.
    """


    params = {
        'api': llm_api,
    }
    params.update(kwargs)
    response = "Temp response"
    return response

def create_session_by_llms_only(llm_api='openai', **kwargs):
    """Creates a new session.

    Args:
        llm_api (str, optional): _description_. Defaults to 'openai'.
    """


    params = {
        'api': llm_api,
    }
    params.update(kwargs)
    response = "Temp response"



import unittest
from unittest.mock import patch
from llm_party.scripts import cli

class TestCLI(unittest.TestCase):
    @patch('llm_party.scripts.cli.start_session_controller')
    def test_start_session_mode(self, mock_start_session_controller):
        # test_args = ['--mode', 'start', '--config_file_path', 'test_config.json']
        test_args = ['dummy_script_name', '--mode', 'start', '--config_file_path', 'test_config.json']
        with patch.object(cli.sys, 'argv', test_args):
            try:
                cli.main()
            except SystemExit as e:
                self.fail(f"SystemExit was raised: {e}")
        mock_start_session_controller.assert_called_once_with('test_config.json', None, None)

    @patch('llm_party.scripts.cli.start_session_with_first_message_controller')
    def test_start_with_first_message_mode(self, mock_start_session_with_first_message_controller):
        test_args = ['dummy_script_name', '--mode', 'start_with_first_message', '--config_file_path', 'test_config.json', '--first_message', 'Hello, world!', '--sender_name', 'Test User', '--sender_role', 'user', '--sender_type', 'Human', '--sender_attendee_params', '{}']
        with patch.object(cli.sys, 'argv', test_args):
            try:
                cli.main()
            except SystemExit as e:
                self.fail(f"SystemExit was raised: {e}")
        mock_start_session_with_first_message_controller.assert_called_once_with('test_config.json', 'Hello, world!', 'Test User', 'user', 'Human', '{}', None, None)

if __name__ == '__main__':
    unittest.main()


import sys
import argparse
from llm_party.controller.session_controller import start_session_controller, start_session_with_first_message_controller

def main():
    parser = argparse.ArgumentParser(description='Start a chat session.')
    parser.add_argument('--mode', choices=['start', 'start_with_first_message'], required=True, help='Mode to run the script in.')
    parser.add_argument('--config_file_path', required=True, help='Path to the configuration file.')
    parser.add_argument('--first_message', help='The first message to send in the chat session.')
    parser.add_argument('--sender_name', help='The name of the sender of the first message.')
    parser.add_argument('--sender_role', help='The role of the sender of the first message.')
    parser.add_argument('--sender_type', help='The type of the sender of the first message.')
    parser.add_argument('--sender_attendee_params', help='Additional parameters for the sender of the first message.')
    parser.add_argument('--save_dir', help='Directory to save the chat session.')
    parser.add_argument('--save_file_name', help='File name to save the chat session.')
    try:
        args = parser.parse_args()
    except SystemExit as e:
        print(f"Error: {e}", file=sys.stderr)
        raise  # Re-raise the exception after printing the error message

    if args.mode == 'start':
        start_session_controller(args.config_file_path, args.save_dir, args.save_file_name)
    elif args.mode == 'start_with_first_message':
        start_session_with_first_message_controller(args.config_file_path, args.first_message, args.sender_name, args.sender_role, args.sender_type, args.sender_attendee_params, args.save_dir, args.save_file_name)

if __name__ == '__main__':
    main()


import unittest
from unittest import skipIf
import json
import yaml
from unittest.mock import patch
from marshmallow import ValidationError
from llm_party.model.session_models import LLMAgent, Message
from llm_party.model.session_models import (
    Instruction,
    Attendee,
    Human,
    LLMAgent,
    Message,
    ChatSession,
    AttendeeSchema,
    MessageSchema,
    ChatSessionSchema
)


USE_ACTUAL_LLM_API = False


def skip_if_no_api():
    return skipIf(
        not USE_ACTUAL_LLM_API, "Skipping this test because USE_ACTUAL_LLM_API is False"
    )


# The return of this method is used a sample JSON to initialize ChatSession objects
def _get_init_json_data():
    return {
        "title": "Chat Session Title",
        "purpose": "Chat Session Purpose",
        "status": "initialized",
        "attendees": [
            {
                "type": "Human",
                "name": "John Doe",
                "role": "user",
                "instruction": {"text": "Say hello"},
                "attendee_params": {},
            },
            {
                "type": "LLMAgent",
                "name": "AI",
                "role": "assistant",
                "instruction": {"text": "Respond politely"},
                "attendee_params": {},
                "llm_api": "openai",
                "llm_api_params": {},
            },
        ],
        "settings": {},
        "chat_history": [],
        "finish_reason": None,
        "start_at": None,
        "updated_at": None,
        "end_at": None,
    }


class TestChatModels(unittest.TestCase):
    def setUp(self):
        # Create a JSON string with the necessary information
        session_init_json = _get_init_json_data()
        self.session = ChatSession(**session_init_json)
        instruction_for_human = Instruction("Say hello")
        instruction_for_agent = Instruction("Respond politely")
        self.instruction = Instruction("Test instruction")
        self.human = Human("John Doe", "user", instruction_for_human)
        self.agent = LLMAgent(
            "AI", "assistant", instruction_for_agent, llm_api="openai"
        )

    def test_instruction(self):
        # Initialize Instruction with some values
        self.instruction = Instruction(
            "Test instruction", "target_value", "version_value", "hash_value"
        )
        self.assertEqual(self.instruction.text, "Test instruction")
        self.assertEqual(self.instruction.target, "target_value")
        self.assertEqual(self.instruction.version, "version_value")
        self.assertEqual(self.instruction.hash, "hash_value")

        # Update the Instruction with new values
        self.instruction.update(
            "Updated instruction", "new_target", "new_version", "new_hash"
        )
        self.assertEqual(self.instruction.text, "Updated instruction")
        self.assertEqual(self.instruction.target, "new_target")
        self.assertEqual(self.instruction.version, "new_version")
        self.assertEqual(self.instruction.hash, "new_hash")

        # Update the Instruction with None values
        self.instruction.update("Another update", None, None, None)
        self.assertEqual(self.instruction.text, "Another update")
        self.assertIsNone(self.instruction.target)
        self.assertIsNone(self.instruction.version)
        self.assertIsNone(self.instruction.hash)

    def test_instruction_with_none_values(self):
        # Create Instruction with None values for target, version, and hash
        instruction = Instruction("Test instruction", None, None, None)
        self.assertEqual(instruction.text, "Test instruction")
        self.assertIsNone(instruction.target, "Target should be None")
        self.assertIsNone(instruction.version, "Version should be None")
        self.assertIsNone(instruction.hash, "Hash should be None")

    def test_attendee(self):
        self.assertEqual(self.human.name, "John Doe")
        self.assertEqual(self.human.role, "user")
        self.assertEqual(self.human.instruction.text, "Say hello")

    def test_chat_session(self):
        self.assertEqual(self.session.title, "Chat Session Title")
        self.assertEqual(self.session.purpose, "Chat Session Purpose")
        self.assertEqual(self.session.status, "initialized")  # Test the initial status
        self.session.start()
        self.assertEqual(
            self.session.status, "started"
        )  # Test the status after starting the session
        self.session.add_attendee(self.human)
        self.assertIn(self.human, self.session.attendees)
        self.session.remove_attendee(self.human)
        self.assertNotIn(self.human, self.session.attendees)

    def test_message(self):
        message = Message(self.human, "Hello, world!")
        self.assertEqual(message.sender, self.human)
        self.assertEqual(message.text, "Hello, world!")

    def test_llm_agent(self):
        self.assertEqual(self.agent.name, "AI")
        self.assertEqual(self.agent.role, "assistant")
        self.assertEqual(self.agent.instruction.text, "Respond politely")

    @skip_if_no_api()
    def test_llm_agent_generate_response(self):
        message = Message(self.human, "Hello, world!")
        response = self.agent.generate_response([message.to_openai_message()])
        self.assertIsNotNone(response)  # Ensure response is not None

    def test_chat_session_to_from_json(self):
        self.session.add_attendee(self.human)
        json_string = self.session.to_json()
        new_session = ChatSession("New Session", "Testing from_json")
        new_session.from_json(json_string)
        self.assertEqual(new_session.title, self.session.title)
        self.assertEqual(new_session.purpose, self.session.purpose)

    def test_chat_session_to_from_json(self):
        self.session.add_attendee(self.human)
        json_string = self.session.to_json()
        new_session = ChatSession.from_json(json_string)
        self.assertEqual(new_session.title, self.session.title)
        self.assertEqual(new_session.purpose, self.session.purpose)

    def test_attendee_send_new_message(self):
        # TODO: Test after implementing send_new_message
        # Human does not support send_new_message yet.
        # message = self.human.send_new_message(self.session)
        pass

    def test_edge_cases(self):
        # Test removing an attendee who is not in the session
        with self.assertRaises(ValueError):
            self.session.remove_attendee(self.agent)

        # Test from_json with invalid JSON string
        with self.assertRaises(ValueError):
            self.session.from_json("invalid json string")

    def test_chat_session_from_json_valid(self):
        valid_json = json.dumps(_get_init_json_data())
        # Use the from_json static method to create a ChatSession instance from the JSON string
        session = ChatSession.from_json(valid_json)
        self.assertEqual(session.title, "Chat Session Title")
        self.assertEqual(session.purpose, "Chat Session Purpose")
        self.assertEqual(len(session.attendees), 2)
        self.assertIsInstance(session.attendees[0], Human)
        self.assertIsInstance(session.attendees[1], LLMAgent)

    def test_remove_attendee_not_in_session(self):
        # Test removing an attendee who is not in the session
        with self.assertRaises(ValueError) as context:
            self.session.remove_attendee(self.agent)
        self.assertIn("Attendee AI is not part of the session", str(context.exception))

    def test_remove_attendee_from_empty_session(self):
        # Test removing an attendee from an empty session
        self.session.attendees = []  # Clear attendees
        with self.assertRaises(ValueError) as context:
            self.session.remove_attendee(self.human)
        self.assertIn(
            "Attendee John Doe is not part of the session", str(context.exception)
        )

    def test_remove_attendee_multiple_times(self):
        # Test removing the same attendee multiple times
        self.session.add_attendee(self.human)
        self.session.remove_attendee(self.human)
        with self.assertRaises(ValueError) as context:
            self.session.remove_attendee(self.human)
        self.assertIn(
            "Attendee John Doe is not part of the session", str(context.exception)
        )

    def test_remove_attendee_not_in_session(self):
        # Test removing an attendee who is not in the session
        with self.assertRaises(ValueError) as context:
            self.session.remove_attendee(self.agent)
        self.assertIn("Attendee AI is not part of the session", str(context.exception))


class TestLLMAgent(unittest.TestCase):
    @patch("llm_party.model.session_models.get_ai_response")
    def test_llm_agent_generate_response(self, mock_get_ai_response):
        # Setup the mock to return a predefined response
        mock_get_ai_response.return_value = "Mocked response"

        instruction_for_agent = Instruction("Respond politely")
        agent = LLMAgent("AI", "assistant", instruction_for_agent, llm_api="openai")
        message = Message(agent, "Hello, world!")
        response = agent.generate_response([message])

        # Assert that the mock was called as expected
        mock_get_ai_response.assert_called_once_with(
            llm_api="openai",
            messages=[
                {"role": "system", "content": agent.instruction.text},
                {"role": "user", "content": message.text},
            ],
            **agent.llm_api_params,
        )
        # Assert that the response is as expected
        self.assertEqual(response, "Mocked response")


class TestChatSessionSerialization(unittest.TestCase):
    def test_chat_session_serialization(self):
        # Given JSON data
        json_data = json.dumps(_get_init_json_data())
        # When deserializing the JSON string to a ChatSession object
        session = ChatSession.from_json(json_data)

        # Then the ChatSession object should have the same data as the JSON
        self.assertEqual(session.title, "Chat Session Title")
        self.assertEqual(session.purpose, "Chat Session Purpose")
        self.assertEqual(len(session.attendees), 2)
        self.assertIsInstance(session.attendees[0], Human)
        self.assertIsInstance(session.attendees[1], LLMAgent)
        self.assertEqual(session.attendees[0].name, "John Doe")
        self.assertEqual(session.attendees[1].name, "AI")
        self.assertEqual(session.attendees[0].instruction.text, "Say hello")
        self.assertEqual(session.attendees[1].instruction.text, "Respond politely")
        self.assertIsNone(session.finish_reason)
        self.assertIsNone(session.start_at)
        self.assertIsNone(session.updated_at)
        self.assertIsNone(session.end_at)


class TestAttendeeSchemaValidation(unittest.TestCase):
    def test_invalid_attendee_type(self):
        invalid_data = {
            "type": "UnknownType",
            "name": "Invalid Attendee",
            "role": "invalid",
            "instruction": {"text": "Invalid instruction"},
        }
        with self.assertRaises(ValidationError):
            AttendeeSchema().load(invalid_data)


class TestChatSessionTimes(unittest.TestCase):
    def test_start_session(self):
        session = ChatSession(title="Test Session", purpose="Testing")
        session.start()
        self.assertIsNotNone(session.start_at)
        self.assertEqual(session.start_at, session.updated_at)


class TestLLMAgentResponseGeneration(unittest.TestCase):
    @patch("llm_party.model.session_models.get_ai_response")
    def test_generate_response(self, mock_get_ai_response):
        mock_get_ai_response.return_value = "Mocked response"
        agent = LLMAgent(
            "AI", "assistant", Instruction("Respond politely"), llm_api="openai"
        )
        message = Message(agent, "Hello, world!")
        response = agent.generate_response([message])
        self.assertEqual(response, "Mocked response")


class TestAttendeeMessageSending(unittest.TestCase):
    def test_send_new_message(self):
        # TODO: Test after implementing send_new_message
        session = ChatSession(title="Test Session", purpose="Testing")
        human = Human("John Doe", "user", Instruction("Say hello"))
        session.add_attendee(human)
        # Human does not support send_new_message yet.
        # human.send_new_message(session)


class TestChatSessionErrorHandling(unittest.TestCase):
    def test_from_json_invalid_json(self):
        with self.assertRaises(ValueError):
            ChatSession.from_json("invalid json string")

    def test_add_existing_attendee(self):
        session = ChatSession(title="Test Session", purpose="Testing")
        human = Human("John Doe", "user", Instruction("Say hello"))
        session.add_attendee(human)
        with self.assertRaises(ValueError):
            session.add_attendee(human)


class TestAttendeeMethods(unittest.TestCase):
    def test_send_new_message(self):
        # Create a Human instance
        human = Human("John", "user", Instruction("Hello"))
        # Create a ChatSession instance
        session = ChatSession("Test Session", "Testing", [human])

        # Test that send_new_message raises a NotImplementedError for Human
        with self.assertRaises(NotImplementedError):
            human.send_new_message(session)

        # Create a LLMAgent instance
        llm_agent = LLMAgent("AI", "bot", Instruction("Hello"), llm_api="openai")
        session.add_attendee(llm_agent)

        # Test that send_new_message works for LLMAgent
        if USE_ACTUAL_LLM_API:
            try:
                message = llm_agent.send_new_message(session)
                self.assertEqual(session.chat_history[-1], message)
            except Exception as e:
                self.fail(f"send_new_message raised {type(e).__name__} unexpectedly!")
        else:
            if False:
                # TODO Fix this block. Mock does not work
                with patch(
                    "llm_wrap.llm_response.get_ai_response",
                    return_value="Mocked response",
                ) as mock_get_ai_response:
                    try:
                        message = llm_agent.send_new_message(session)
                        self.assertEqual(session.chat_history[-1], message)
                    except Exception as e:
                        self.fail(
                            f"send_new_message raised {type(e).__name__} unexpectedly!"
                        )

                    # Check that the mock was called
                    mock_get_ai_response.assert_called_once()


class TestLLMAgentMethods(unittest.TestCase):
    def test_convert_chat_history_to_openai_messages(self):
        # Setup
        llm_agent = LLMAgent("AI", "bot", Instruction("Hello"), llm_api="openai")
        human = Human("John", "user", Instruction("Hi"))
        chat_history = [
            Message(human, "Hi"),
            Message(llm_agent, "Hello"),
            Message(human, "How are you?"),
        ]

        # Execution
        messages = llm_agent.convert_chat_history_to_openai_messages(chat_history)

        # Verification
        self.assertEqual(
            len(messages), len(chat_history) + 1
        )  # +1 for the system message
        self.assertEqual(
            messages[0], {"role": "system", "content": llm_agent.instruction.text}
        )
        for i, message in enumerate(chat_history, start=1):
            expected_role = "user" if i % 2 == 1 else "assistant"
            self.assertEqual(
                messages[i], {"role": expected_role, "content": message.text}
            )


class TestChatSessionEndMethod(unittest.TestCase):
    def setUp(self):
        # Create a ChatSession instance
        session_init_json = _get_init_json_data()
        self.session = ChatSession(**session_init_json)

    def test_end_session(self):
        # Start the session
        self.session.start()
        self.assertEqual(self.session.status, "started")

        # End the session successfully
        self.session.end("completed_successfully", "MaxSentMessage")
        self.assertEqual(self.session.status, "completed_successfully")
        self.assertEqual(self.session.finish_reason, "MaxSentMessage")
        self.assertIsNotNone(self.session.end_at)

        # End the session with failure
        self.session.end("completed_with_failure", "LLMAPIFailure")
        self.assertEqual(self.session.status, "completed_with_failure")
        self.assertEqual(self.session.finish_reason, "LLMAPIFailure")
        self.assertIsNotNone(self.session.end_at)


class TestMessageSchema(unittest.TestCase):
    def setUp(self):
        self.message_schema = MessageSchema()
        self.attendee_schema = AttendeeSchema()
        self.instruction = Instruction("Test instruction")
        self.attendee = Human(
            "Test Attendee", "user", self.instruction
        )
        self.message = Message(self.attendee, "Test message")

    def test_serialize_message(self):
        serialized_message = self.message_schema.dump(self.message)
        expected_output = {
            "sender": self.attendee_schema.dump(self.attendee),
            "text": "Test message",
            "timestamp": self.message.timestamp.strftime("%Y-%m-%dT%H:%M:%S"),
        }
        self.assertEqual(serialized_message, expected_output)

    def test_deserialize_message(self):
        serialized_message = self.message_schema.dump(self.message)
        deserialized_message = self.message_schema.load(serialized_message)
        self.assertEqual(deserialized_message.sender.name, self.message.sender.name)
        self.assertEqual(deserialized_message.text, self.message.text)
        # Compare timestamps up to seconds
        self.assertEqual(
            deserialized_message.timestamp.replace(microsecond=0),
            self.message.timestamp.replace(microsecond=0),
        )

    def test_deserialize_invalid_message(self):
        invalid_message = {
            "sender": self.attendee_schema.dump(self.attendee),
            "text": "Test message",
            "timestamp": "invalid timestamp",
        }
        with self.assertRaises(ValidationError):
            self.message_schema.load(invalid_message)


class TestChatSessionSchema(unittest.TestCase):
    def setUp(self):
        self.chat_session_schema = ChatSessionSchema()
        self.attendee_schema = AttendeeSchema()
        self.message_schema = MessageSchema()
        self.instruction = Instruction("Test instruction")
        self.attendee = Human(
            "Test Attendee", "user", self.instruction
        )
        self.message = Message(self.attendee, "Test message")
        self.chat_session = ChatSession(
            title="Test Session",
            purpose="Testing",
            attendees=[self.attendee],
            chat_history=[self.message]
        )

    def test_serialize_chat_session(self):
        serialized_chat_session = self.chat_session_schema.dump(self.chat_session)
        expected_output = {
            'title': "Test Session",
            'purpose': "Testing",
            'attendees': [self.attendee_schema.dump(self.attendee)],
            'settings': {},
            'start_at': None,
            'updated_at': None,
            'end_at': None,
            'finish_reason': None,
            'status': "initialized",
            'chat_history': [self.message_schema.dump(self.message)]
        }
        self.assertEqual(serialized_chat_session, expected_output)

    def test_deserialize_chat_session(self):
        serialized_chat_session = self.chat_session_schema.dump(self.chat_session)
        deserialized_chat_session = self.chat_session_schema.load(serialized_chat_session)
        self.assertEqual(deserialized_chat_session.title, self.chat_session.title)
        self.assertEqual(deserialized_chat_session.purpose, self.chat_session.purpose)
        self.assertEqual(deserialized_chat_session.attendees[0].name, self.chat_session.attendees[0].name)
        self.assertEqual(deserialized_chat_session.chat_history[0].text, self.chat_session.chat_history[0].text)

    def test_deserialize_invalid_chat_session(self):
        invalid_chat_session = {
            'purpose': "Testing",
            'attendees': [self.attendee_schema.dump(self.attendee)],
            'settings': {},
            'start_at': None,
            'updated_at': None,
            'end_at': None,
            'finish_reason': None,
            'status': "initialized",
            'chat_history': [self.message_schema.dump(self.message)]
        }
        with self.assertRaises(ValidationError):
            self.chat_session_schema.load(invalid_chat_session)

class TestChatSessionFromJsonToJson(unittest.TestCase):
    def setUp(self):
        self.chat_session_schema = ChatSessionSchema()
        self.attendee_schema = AttendeeSchema()
        self.message_schema = MessageSchema()
        self.instruction = Instruction("Test instruction")
        self.attendee = Human(
            "Test Attendee", "user", self.instruction
        )
        self.message = Message(self.attendee, "Test message")
        self.chat_session = ChatSession(
            title="Test Session",
            purpose="Testing",
            attendees=[self.attendee],
            chat_history=[self.message]
        )

    def test_to_json(self):
        serialized_chat_session = self.chat_session.to_json()
        expected_output = self.chat_session_schema.dumps(self.chat_session)
        self.assertEqual(serialized_chat_session, expected_output)

    def test_from_json(self):
        serialized_chat_session = self.chat_session.to_json()
        deserialized_chat_session = ChatSession.from_json(serialized_chat_session)
        self.assertEqual(deserialized_chat_session.title, self.chat_session.title)
        self.assertEqual(deserialized_chat_session.purpose, self.chat_session.purpose)
        self.assertEqual(deserialized_chat_session.attendees[0].name, self.chat_session.attendees[0].name)
        self.assertEqual(deserialized_chat_session.chat_history[0].text, self.chat_session.chat_history[0].text)

    def test_to_dict(self):
        dict_data = self.chat_session.to_dict()
        expected_output = self.chat_session_schema.dumps(self.chat_session)
        self.assertEqual(dict_data, yaml.safe_load(expected_output))

    def test_to_yaml(self):
        yaml_data = self.chat_session.to_yaml()
        expected_output = yaml.dump(yaml.safe_load(self.chat_session_schema.dumps(self.chat_session)), default_flow_style=False)
        self.assertEqual(yaml_data, expected_output)

class TestMessage(unittest.TestCase):
    def test_create(self):
        # Test with a Human sender
        sender_name = "Test Human"
        sender_role = "user"
        sender_type = "Human"
        sender_attendee_params = {}
        text = "Hello, world!"
        message = Message.create(sender_name, sender_role, sender_type, sender_attendee_params, text)
        self.assertIsInstance(message.sender, Human)
        self.assertEqual(message.text, text)

        # Test with an LLMAgent sender
        sender_name = "Test LLMAgent"
        sender_role = "assistant"
        sender_type = "LLMAgent"
        sender_attendee_params = {"llm_api": "openai"}
        text = "Hello, user!"
        message = Message.create(sender_name, sender_role, sender_type, sender_attendee_params, text)
        self.assertIsInstance(message.sender, LLMAgent)
        self.assertEqual(message.text, text)

        # Test with an invalid sender_type
        sender_type = "InvalidType"
        with self.assertRaises(ValueError):
            message = Message.create(sender_name, sender_role, sender_type, sender_attendee_params, text)

class TestChatSession(unittest.TestCase):
    def setUp(self):
        self.instruction = Instruction("Test instruction")
        self.attendee = Human(
            "Test Attendee", "user", self.instruction
        )
        self.message = Message(self.attendee, "Test message")
        self.chat_session = ChatSession(
            title="Test Session",
            purpose="Testing",
            attendees=[self.attendee],
            chat_history=[self.message]
        )

    def test_get_message_from_history(self):
        message_dict = self.chat_session.get_message_from_history(0)
        self.assertIsInstance(message_dict, dict)
        self.assertEqual(message_dict["sender"]["name"], self.message.sender.name)
        self.assertEqual(message_dict["text"], self.message.text)

class TestChatSessionSetMaxMessages(unittest.TestCase):
    def setUp(self):
        self.chat_session = ChatSession(
            title="Test Session",
            purpose="Testing",
            attendees=[],
            chat_history=[]
        )

    def test_set_max_messages(self):
        # Set max_sent_message to a specific value
        self.chat_session.set_max_messages(10)
        self.assertEqual(self.chat_session.settings["max_sent_message"], 10)

        # Change max_sent_message to a new value
        self.chat_session.set_max_messages(20)
        self.assertEqual(self.chat_session.settings["max_sent_message"], 20)

if __name__ == '__main__':
    unittest.main()




from setuptools import setup, find_packages

setup(
    name='llm_party',
    version='0.1.4',
    author='Akihito Sudo',
    author_email="akihito.s.gm@gmail.com",
    url="https://github.com/sudodo/llm_party",
    packages=find_packages(),
    install_requires=[
        'llm_wrap==0.1.2',
        'marshmallow==3.20.1',
        'PyYAML==6.0.1',
    ],
    include_package_data=True,
)



from typing import Union, Dict, Optional, Callable
from functools import wraps
import yaml
import os
import json
import tempfile
from typing import Any, Dict, Tuple, Literal
from llm_party.service.session_service import (
    start_session,
    start_session_with_user_first_message,
    initialize_chat_session
)

from ..model.session_models import ChatSession


def is_yaml(file_path):
    return file_path.endswith(".yaml") or file_path.endswith(".yml")


def convert_yaml_to_temp_json(yaml_file_path: str) -> str:
    """Converts a YAML file to a temporary JSON file. The temporary file will be deleted after the program exits.

    Args:
        yaml_file_path (str): Yaml file path

    Returns:
        str: Json file path of the converted yaml file.
    """
    with open(yaml_file_path, "r") as file:
        config_dict = yaml.safe_load(file)

    temp_file = tempfile.NamedTemporaryFile(mode="w+", delete=False, suffix=".json")
    json.dump(config_dict, temp_file)
    temp_file.flush()

    return temp_file.name


def yaml_acceptable(func):
    """
    A decorator that preprocesses the 'config_file_path' argument for the decorated function.

    If the 'config_file_path' argument points to a YAML file, this decorator converts the file to a temporary JSON format before calling the original function. The conversion is performed using the 'convert_yaml_to_temp_json' function, and the check for YAML file format is done using the 'is_yaml' function.

    Parameters:
    func (Callable): The function to be decorated. It is expected that this function will have 'config_file_path' as one of its arguments.

    Returns:
    Callable: A wrapper function that incorporates the preprocessing step and then calls the original function with the updated arguments.

    Usage:
    @preprocess_config
    def my_function(config_file_path, *args, **kwargs):
        # Function implementation
        pass
    """

    def wrapper(*args, **kwargs):
        # Extract and preprocess config_file_path
        config_file_path = args[0] if args else kwargs.get("config_file_path")
        is_yaml_config = is_yaml(config_file_path)
        if config_file_path and is_yaml_config:
            json_config_file_path = convert_yaml_to_temp_json(config_file_path)
        else:
            json_config_file_path = config_file_path

        if args:
            args = (json_config_file_path,) + args[1:]
        else:
            kwargs["config_file_path"] = json_config_file_path

        try:
            # Call the original function
            ret = func(*args, **kwargs)
        finally:
            if is_yaml_config:
                os.remove(json_config_file_path)
        return ret

    return wrapper

def config_dict_or_path(func):
    """
    A decorator that preprocesses the 'config' argument for the decorated function.

    If the 'config' argument is a dictionary, this decorator converts the dictionary to a temporary JSON file before calling the original function. If the 'config' argument is a string, it is assumed to be a file path and is passed directly to the original function.

    Parameters:
    func (Callable): The function to be decorated. It is expected that this function will have 'config' as one of its arguments.

    Returns:
    Callable: A wrapper function that incorporates the preprocessing step and then calls the original function with the updated arguments.
    """
    @wraps(func)
    def wrapper(config: Union[str, Dict], *args, **kwargs):
        if isinstance(config, dict):
            temp_file = tempfile.NamedTemporaryFile(mode="w+", delete=False, suffix=".json")
            json.dump(config, temp_file)
            temp_file.flush()
            config_file_path = temp_file.name
        else:
            config_file_path = config

        try:
            # Call the original function
            ret = func(config_file_path, *args, **kwargs)
        finally:
            if isinstance(config, dict):
                os.remove(config_file_path)
        return ret

    return wrapper

def config_chat_session_or_dict_or_path_or_yaml(func):
    """
    Decorator function that handles the configuration of a chat session.
    It accepts a configuration parameter that can be a string, a dictionary, or a ChatSession object.
    If the configuration is a ChatSession object, it is converted to a temporary JSON file.
    If the configuration is a dictionary, it is also converted to a temporary JSON file.
    If the configuration is a YAML file, it is converted to a temporary JSON file.
    Otherwise, the configuration is treated as a file path.

    Args:
        func (callable): The function to be decorated.

    Returns:
        callable: The decorated function.

    Raises:
        FileNotFoundError: If the configuration file path is not found.
    """
    @wraps(func)
    def wrapper(config: Union[str, Dict, ChatSession], *args, **kwargs):
        if isinstance(config, ChatSession):
            temp_file = tempfile.NamedTemporaryFile(mode="w+", delete=False, suffix=".json")
            json.dump(config.to_dict(), temp_file)
            temp_file.flush()
            config_file_path = temp_file.name
        elif isinstance(config, dict):
            temp_file = tempfile.NamedTemporaryFile(mode="w+", delete=False, suffix=".json")
            json.dump(config, temp_file)
            temp_file.flush()
            config_file_path = temp_file.name
        else:
            if is_yaml(config):
                config_file_path = convert_yaml_to_temp_json(config)
            else:
                config_file_path = config

        try:
            # Call the original function
            ret = func(config_file_path, *args, **kwargs)
        finally:
            if isinstance(config, (dict, ChatSession)) or is_yaml(config):
                os.remove(config_file_path)
        return ret

    return wrapper

@config_chat_session_or_dict_or_path_or_yaml
def init_chat_session(config: Union[str, Dict, ChatSession]):
    return initialize_chat_session(config)

@config_chat_session_or_dict_or_path_or_yaml
def start_session_controller(config: Union[str, Dict, ChatSession], save_dir:str=None, save_file_name:str=None, pretty_print:bool=False, save_yaml:bool=False, output: Optional[Callable[[str], None]] = None):
    chat_session, saved_file_path = start_session(config, save_dir, save_file_name, pretty_print, save_yaml, output)

    return chat_session, saved_file_path

@config_chat_session_or_dict_or_path_or_yaml
def start_session_with_first_message_controller(
    config: Union[str, Dict, ChatSession],
    first_message: str,
    sender_name: str,
    sender_role: str,
    sender_type: Literal["Human", "Agent"],
    sender_attendee_params: dict,
    save_dir:str=None,
    save_file_name: str=None,
    pretty_print:bool=False,
    save_yaml:bool=False,
    output: Optional[Callable[[str], None]] = None
) -> Tuple[ChatSession, str]:
    chat_session, saved_file_path = start_session_with_user_first_message(
        config,
        first_message,
        sender_name,
        sender_role,
        sender_type,
        sender_attendee_params,
        save_dir,
        save_file_name,
        pretty_print,
        save_yaml,
        output
    )

    return chat_session, saved_file_path



from llm_party.attendee.default.default_attendee_trait import FacilitatorTrait
from llm_party import RawLLMAgent
from llm_party import AttendeeTraitBase
from typing import Final, List

class GPTStaffTrait(AttendeeTraitBase):
    character_list: Final[List[str]] = [
        "gpt_staff"
    ]

class GPTProductOwner(RawLLMAgent):
    """
    This role is for the product owner of the GPT project.
    Because it has the FacilitatorTrait, it can terminate the session.
    """
    character_classes = [FacilitatorTrait, GPTStaffTrait]

class GPTInnovativeManager(RawLLMAgent):
    pass
